import cPickle


class Packet(object):

    def serialize(self):
        return cPickle.dumps(self)


class PacketResult(Packet):

    def __init__(self, success, message=None):
        self.success = success
        if message is None and success is False:
            message = "Something went wrong"
        if message is None and success is True:
            message = ''
        self.message = message


class PacketSet(Packet):

    def __init__(self, packets):
        self.packets = packets


class PlayerQuitPacket(Packet):
    def __init__(self):
        super(PlayerQuitPacket, self)


from LoreleiLib.Packets.Common import Packet, PacketResult


# Server Data
class CreationOptionsResponse(Packet):

    def __init__(self, races, classes):
        self.races = races
        self.classes = classes


class CharacterCheckResponse(Packet):

    def __init__(self, character):
        self.character = character


class CharacterFinishResponse(PacketResult):
    def __init__(self, success, message):
        super(CharacterFinishResponse, self).__init__(success, message)


# Client Requests
class OptionsRequest(Packet):

    # Packet from client to request Race/Class options
    def __init__(self):
        pass


class CharacterSummaryBackRequest(Packet):
    def __init__(self):
        pass


class RaceSelectionRequest(Packet):

    # Packet from client to set Race on the Server
    def __init__(self, characterRace):
        self.characterRace = characterRace


class ClassSelectionRequest(Packet):

    # Packet from client to set Class on the Server
    def __init__(self, characterClass):
        self.characterClass = characterClass


class CharacterCheckRequest(Packet):

    def __init__(self):
        pass


class CharacterFinishRequest(Packet):

    def __init__(self):
        pass


from enum import Enum
from LoreleiLib.Helpers import firstLetterLowercase
from collections import OrderedDict


# If you add a value to these enums
# Add the matching camel cased name to Attributes for type intellisense
class Vital(Enum):
    Health = 1
    Mana = 2
    Stamina = 3


class Attribute(Enum):
    Strength = 1
    Constitution = 2
    Dexterity = 3
    Agility = 4
    Wisdom = 5
    Intelligence = 6
    Charisma = 7


class Statistic(Enum):
    Attack = 1
    MagicAttack = 2
    Defense = 3
    MagicDefense = 4
    Accuracy = 5
    MagicAccuracy = 6
    Dodge = 7
    MagicDodge = 8
    Critical = 9
    Threat = 10
    AttackSpeed = 11
    CastSpeed = 12


class AttributeSet(object):

    def __init__(self, **kwargs):
        # Vitals
        self.health = 0
        self.maxHealth = 0
        self.mana = 0
        self.maxMana = 0
        self.stamina = 0
        self.maxStamina = 0

        # Attributes
        self.strength = 0
        self.constitution = 0
        self.dexterity = 0
        self.agility = 0
        self.wisdom = 0
        self.intelligence = 0
        self.charisma = 0

        # Stats
        self.attack = 0
        self.magicAttack = 0
        self.defense = 0
        self.magicDefense = 0
        self.accuracy = 0
        self.magicAccuracy = 0
        self.dodge = 0
        self.magicDodge = 0
        self.critical = 0
        self.magicCritical = 0
        self.threat = 0
        self.attackSpeed = 0.0
        self.castSpeed = 0.0

        allowed = []
        for vital in Vital:
            allowed.append(vital.name)
        for attribute in Attribute:
            allowed.append(attribute.name)
        for stat in Statistic:
            allowed.append(stat.name)

        for key, value in kwargs.iteritems():
            if key in allowed:
                self.__setattr__(firstLetterLowercase(key), value)

    def getDict(self):
        attributeDict = OrderedDict()
        # Vitals
        attributeDict["Max Health"] = self.maxHealth
        attributeDict["Max Mana"] = self.maxMana
        attributeDict["Max Stamina"] = self.maxStamina

        # Physical
        attributeDict["P. Attack"] = self.attack
        attributeDict["P. Defense"] = self.defense
        attributeDict["Accuracy"] = self.accuracy
        attributeDict["Dodge"] = self.dodge
        attributeDict["Critical"] = self.critical
        attributeDict["Attack Speed"] = self.attackSpeed

        # Magical
        attributeDict["M. Attack"] = self.magicAttack
        attributeDict["M. Defense"] = self.magicDefense
        attributeDict["M. Accuracy"] = self.magicAccuracy
        attributeDict["Magic Dodge"] = self.magicDodge
        attributeDict["Magic Critical"] = self.magicCritical
        attributeDict["Cast Speed"] = self.castSpeed

        # Other
        attributeDict["Threat"] = self.threat
        return attributeDict

    def getViewDict(self):
        attributeDict = OrderedDict()
        # Vitals
        attributeDict["Vitals"] = OrderedDict()
        attributeDict["Vitals"]["Max Health"] = self.maxHealth
        attributeDict["Vitals"]["Max Mana"] = self.maxMana
        attributeDict["Vitals"]["Max Stamina"] = self.maxStamina

        # Physical
        attributeDict["Physical"] = OrderedDict()
        attributeDict["Physical"]["P. Attack"] = self.attack
        attributeDict["Physical"]["P. Defense"] = self.defense
        attributeDict["Physical"]["Accuracy"] = self.accuracy
        attributeDict["Physical"]["Dodge"] = self.dodge
        attributeDict["Physical"]["Critical"] = self.critical
        attributeDict["Physical"]["Attack Speed"] = self.attackSpeed

        # Magical
        attributeDict["Magical"] = OrderedDict()
        attributeDict["Magical"]["M. Attack"] = self.magicAttack
        attributeDict["Magical"]["M. Defense"] = self.magicDefense
        attributeDict["Magical"]["M. Accuracy"] = self.magicAccuracy
        attributeDict["Magical"]["M. Dodge"] = self.magicDodge
        attributeDict["Magical"]["M. Critical"] = self.magicCritical
        attributeDict["Magical"]["Cast Speed"] = self.castSpeed

        # Other
        attributeDict["Other"] = OrderedDict()
        attributeDict["Other"]["Enmity"] = self.threat
        return attributeDict

    def getNonZeroDict(self):
        attributeDict = OrderedDict()
        if self.maxHealth is not 0 and self.maxHealth is not None:
            attributeDict["Max Health"] = self.maxHealth
        if self.maxMana is not 0 and self.maxMana is not None:
            attributeDict["Max Mana"] = self.maxMana
        if self.maxStamina is not 0 and self.maxStamina is not None:
            attributeDict["Max Stamina"] = self.maxStamina

        if self.attack is not 0 and self.attack is not None:
            attributeDict["P. Attack"] = self.attack
        if self.defense is not 0 and self.defense is not None:
            attributeDict["P. Defense"] = self.defense
        if self.accuracy is not 0 and self.accuracy is not None:
            attributeDict["Accuracy"] = self.accuracy
        if self.dodge is not 0 and self.dodge is not None:
            attributeDict["Dodge"] = self.dodge
        if self.critical is not 0 and self.critical is not None:
            attributeDict["Critical"] = self.critical
        if int(self.attackSpeed) is not 0 and self.attackSpeed is not None and self.attackSpeed is not 0.0:
            attributeDict["Attack Speed"] = self.attackSpeed

        if self.magicAttack is not 0 and self.magicAttack is not None:
            attributeDict["M. Attack"] = self.magicAttack
        if self.magicDefense is not 0 and self.magicDefense is not None:
            attributeDict["M. Defense"] = self.magicDefense
        if self.magicDodge is not 0 and self.magicDodge is not None:
            attributeDict["M. Dodge"] = self.magicDodge
        if self.magicCritical is not 0 and self.magicCritical is not None:
            attributeDict["M. Critical"] = self.magicCritical
        if int(self.castSpeed) is not 0 and self.castSpeed is not None and self.castSpeed is not 0.0:
            attributeDict["Cast Speed"] = self.castSpeed

        if self.threat is not 0 and self.threat is not None:
            attributeDict["Enmity"] = self.threat

        return attributeDict

    def setValues(self, otherSet):
        # type: (AttributeSet) -> AttributeSet
        for key, value in vars(otherSet):
            self.__setattr__(key, value)



from LoreleiLib.Packets.Common import Packet
from enum import Enum


class MessageType(Enum):

    # Person messages
    Global = 0
    Region = 1
    Local = 2 # Room
    Whisper = 3

    # System messages
    Combat = 4
    System = 5


class ChatPacket(Packet): pass


# Client message sent packet
class MessageSent(ChatPacket):

    def __init__(self, type, message, target=None):
        super(MessageSent)
        self.type = type
        self.message = message
        self.target = target # Target player


class MessageRecieved(ChatPacket):

    def __init__(self, type, message, sender=None, target=None):
        self.type = type
        self.message = message
        self.sender = sender
        self.target = target

    def __str__(self):
        if self.type in [MessageType.Combat, MessageType.System]:
            return "[{}] : {}".format(self.type.name, self.message)
        elif self.type in [MessageType.Whisper]:
            return "[{}] >> {}".format(self.sender, self.message)
        else:
            return "[{}] : {}".format(self.sender, self.message)


from LoreleiLib.Statistics import AttributeSet
from LoreleiLib.Objects import Living, Equipment, ArmorType, WeaponType, Item, EquipmentSlot, Weapon
from LoreleiLib.CombatHelpers import CalculateNextLevelExperience, CalculateExperienceGain


class CharacterClass(object):

    def __init__(self, name, description, skillList, weaponList, armorList, starterSet=None,
                 startingAttributes=AttributeSet(Strength=10, Constitution=10, Dexterity=10, Agility=10, Intelligence=10, Wisdom=10, Charisma=10)):
        # type: (str, str, dict, list, list, list, AttributeSet, dict) -> None
        self.name = name # type: str
        self.description = description # type: str
        self.skillList = skillList # type: dict<int, list>
        self.weapons = weaponList # type: list<WeaponType>
        self.armor = armorList # type: list<ArmorType>
        self.startingAttributes = startingAttributes # type: AttributeSet
        self.starterSet = starterSet # type: list<str>
        self.starterEquipment = [] # type: dict<EquipmentSlot, Item>

    def getFullDescription(self):
        descrip = self.description + "\n\n"
        descrip += "Weapons : "
        for weaponType in self.weapons:
            descrip += weaponType.name.replace("_", " ") + ", "
        descrip = descrip[:-2]
        descrip += "\n\n"
        descrip += "Armor : "
        for armorType in self.armor:
            descrip += armorType.name.replace("_", " ") + ", "
        descrip = descrip[:-2]
        return descrip


class CharacterRace(object):

    def __init__(self, name, description, attributes):
        # type: (str, str, AttributeSet) -> None
        self.name = name # type: str
        self.description = description # type: str
        self.attributes = attributes # type: AttributeSet


class Character(Living):

    def __init__(self, name, character_race, character_class, level=1, experience=0):
        self.name = name # type: str
        self.character_race = character_race # type: CharacterRace
        self.character_class = character_class # type: CharacterClass
        self.inventory = [] # type: list

        super(Character, self).__init__(name, "A random traveller", level, AttributeSet())
        self.experience = experience # type: int
        self.nextLevel = CalculateNextLevelExperience(level)# type: int
        self.equipment = Equipment(self) # type: Equipment

        # Core + Equipment = Stats used for battle
        self.coreAttributes = AttributeSet(Strength=0, Constitution=0, Dexterity=0, Agility=0, Wisdom=0, Intelligence=0, Charisma=0) # type: AttributeSet
        self.buffs = [] # type: list

    def buildStats(self):
        self.attributes = AttributeSet()
        self.buildCoreAttributes()
        self.buildEquipmentAttributes()
        self.makeStatistics()

    def buildCoreAttributes(self):
        # Use alloted attributes
        for key, value in self.coreAttributes.__dict__.iteritems():
            setattr(self.attributes, key, value)

        # Add race attributes
        if self.character_race is not None:
            for key, value in self.character_race.attributes.__dict__.iteritems():
                setattr(self.attributes, key, value)

        # Add starting class attributes
        if self.character_class is not None:
            for key, value in self.character_class.startingAttributes.__dict__.iteritems():
                setattr(self.attributes, key, getattr(self.attributes, key) + value)

    def buildEquipmentAttributes(self):
        for key, value in self.equipment.__dict__.iteritems():
            if isinstance(value, Weapon):
                if key == "main_hand":
                    self.attributes.attackSpeed += value.attackSpeed
                elif key == "off_hand":
                    self.attributes.attackSpeed += (value.attackSpeed / 2)
            if key != "owner" and value is not None:
                for key2, value2 in value.attributes.__dict__.iteritems():
                    setattr(self.attributes, key2, getattr(self.attributes, key2) + value2)

    def makeStatistics(self):
        self.attributes.maxHealth += 10 + self.attributes.strength + self.attributes.constitution * 3
        self.attributes.maxMana += 5 + (self.attributes.intelligence +  self.attributes.wisdom * 3) / 2
        self.attributes.maxStamina += 5 + (self.attributes.strength * 2 + self.attributes.constitution * 2) / 2

        self.attributes.attack += int(self.attributes.strength * 1.5)
        self.attributes.defense += self.attributes.constitution * 2
        self.attributes.accuracy += 80 + self.attributes.dexterity / 3
        self.attributes.dodge += 1 + (self.attributes.agility-5) / 5
        self.attributes.critical += 1 + ((self.attributes.agility-5) / 10) + ((self.attributes.dexterity-5) / 10) - int(self.level / 10)
        self.attributes.attackSpeed -= 0.0 # - 0 TODO: Make formula

        self.attributes.magicAttack += int(self.attributes.intelligence * 1.5)
        self.attributes.magicDefense += self.attributes.wisdom * 2
        self.attributes.magicAccuracy += 80 + self.attributes.intelligence / 3
        self.attributes.magicDodge += 1 # - 0 TODO: Make formula
        self.attributes.magicCritical += (self.attributes.intelligence + self.attributes.wisdom) / 10 - int(self.level / 10)
        self.attributes.castSpeed = 1.0 # - 0 TODO: Make formula

        threatValue = int(self.attributes.constitution / 10 - 1)
        if threatValue < 0:
            threatValue = 0
        self.attributes.threat += threatValue

    def changeRace(self, race):
        # type: (CharacterRace) -> None
        self.character_race = race

    def giveReward(self, target):
        expGain = CalculateExperienceGain(target.level, self.level)
        if (expGain > 0):
            self.experience += expGain
            if self.experience > self.nextLevel:
                self.levelUp()

    def levelUp(self):
        self.level += 1
        self.experience -= self.nextLevel
        self.nextLevel = CalculateNextLevelExperience(self.level)

    def damage(self, amount, attacker):
        if attacker is not None:
            self.send('{} dealt {} damage to you'.format(attacker.name, amount))
            super(Character, self).damage(amount, attacker)

    def send(self, message):
        raise NotImplementedError

    def giveItem(self, item, announce=True):
        if item is not None:
            self.inventory.append(item)
            if announce:
                self.send("You have received a(n) {}".format(item.name))

    def equip(self, item, slot):
        unequipped = self.equipment.equip(item, slot)
        for item in unequipped:
            self.inventory.append(item)

    def unequip(self, slot):
        item = self.equipment.unequip(slot)
        if item is not None:
            self.inventory.append(item)


from setuptools import setup, find_packages


setup(name='LoreleiLib',
    version='0.1.7.4', # Don't forget to make a changelog entry in README.rst
    description='The text based MUD made by Lorinthio',
    author='Lorinthio',
    author_email='benrandallswg@gmail.com',
    packages=find_packages(),
    install_requires=[
      'twisted',
      'pygame',
      'enum34'
    ],
    zip_safe=False)


from LoreleiLib.Packets.Common import Packet, PacketResult


# Server Data
class LoginAttemptResponse(PacketResult):

    def __init__(self, success, message=None):
        super(LoginAttemptResponse, self).__init__(success, message)


class AccountCreateResponse(PacketResult):
    def __init__(self, success, message=None):
        super(AccountCreateResponse, self).__init__(success, message)


# Client Requests
class LoginAttemptRequest(Packet):

    def __init__(self, username, password):
        self.username = username
        self.password = password


class AccountCreateRequest(Packet):

    def __init__(self, username, password):
        self.username = username
        self.password = password






from itertools import chain


def truncline(text, font, maxwidth):
    real = len(text)
    stext = text
    l = font.size(text)[0]
    cut = 0
    a = 0
    done = 1
    old = None
    while l > maxwidth:
        a = a + 1
        n = text.rsplit(None, a)[0]
        if stext == n:
            cut += 1
            stext = n[:-cut]
        else:
            stext = n
        l = font.size(stext)[0]
        real = len(stext)
        done = 0
    return real, done, stext


def wrapline(text, font, maxwidth):
    done = 0
    wrapped = []

    while not done:
        nl, done, stext = truncline(text, font, maxwidth)
        wrapped.append(stext.strip())
        text = text[nl:]
    return wrapped


def wrap_multi_line(text, font, maxwidth):
    """ returns text taking new lines into account.
    """
    lines = chain(*(wrapline(line, font, maxwidth) for line in text.splitlines()))
    return list(lines)